
//*********************************************************************************************************
//   Program written for Arduino Nano to drive Vishay "AGC P" type proximity sensor consisting of TSSP4P38 
//   receiver (connected to pin D8) and VSLB3940 IR emitter (connected with a series resistor to pin D3). Best 
//   results are obtained with a burst length between 100ms and 150ms and a burst period of 500ms or more. The
//   TSSP4P38 requires a long idle period for its AGC to recover full sensitivity after repressing the burst.
//   The serial monitor must be used at 9600 baud and send Newline characters to enter the sensor parameters.
//   Authors: Anika Kühnle & John Fisher
//*********************************************************************************************************
const int BUFFER_SIZE = 4;                  // buffer for serial input, max 3 digits for burst and burstrep
uint16_t burst = 0;                         // length of burst in ms
uint16_t burstrep = 0;                      // burst period in ms
uint16_t burst_count = 0;                   // number of carrier cycles in burst
uint16_t burstrep_count = 0;                // number of carrier cycles in burst period
volatile uint16_t input_capture =0;         // counter used to capture the receiver pulse length in timer1
enum {INACTIVE, ACTIVE} state = INACTIVE;   // INACTIVE is waiting for burst times, ACTIVE is where Rx/TX is started or stopped
enum {PENDING, DONE, BAD} rx_flag = PENDING;// PENDING= Rx is capturing a pulse, DONE=Rx captured the pulse
const uint8_t HALF_PERIOD = 210;            // (Period_38kHz/2) / Period_16MHz
const uint8_t DUTY = 63;                    // duty cycle of the 38kHz signal = 30% x HALF_PERIOD (e.g. 30% duty cycle, 30% x HALF_PERIOD = 630)

//*********************************************************************************************************
//   Function to input a line of alpha numeric characters. Requires an array and its size.
//   Returns the number of characters read.
//*********************************************************************************************************
int getline(char* buf, int bufsize){
  for (int bufctr = 0; bufctr < bufsize; bufctr++) {
    while (Serial.available() == 0) { }     // do nothing if no chars received yet
    char onechar= Serial.read();            // read one character
    if (onechar== '\n') {                   // if EOL
      buf[bufctr] = 0;                      // then add string terminator
      return bufctr;                        // and return length of string
    }                                       // end if
    else {
      buf[bufctr] = onechar;                // else add character to buf
    }                                       // end else
  }                                         // end for

  // If no EOL was received, issue warning
  Serial.println("Warning, either too many characters were sent, or");
  Serial.println("the Serial Monitor is not set to send Newline characters.");
  Serial.println("Please correct, and restart the monitor.");
  buf[0] = 0;                               // discard the input
  return -1;                                // return error
}                                           // end getline

//*********************************************************************************************************
//   Function to print a header.
//*********************************************************************************************************
void print_header() {
    Serial.println("*PLEASE ENTER THE SIGNAL SETTINGS*");
    Serial.println("(Press key '1' for start sending, key '0' for stop sending)");
    Serial.println(" ");
    Serial.print("Length of Burst (ms):");
    Serial.print("\t\t");
    Serial.print("Burst Repetition Time (ms):");
    Serial.println(" ");    
}                                           // end print_header

//*********************************************************************************************************
//  Interrupt service routine
//  Timer2 overflow - used to count carrier cycles
//*********************************************************************************************************
ISR(TIMER2_OVF_vect){
  static uint16_t carrier_count = 0;       // persistent counter used to count carrier cycles
  carrier_count++;                         // increment carrier counter
  if (carrier_count == burst_count){       // if burst is completed
    TCCR2A &= ~(1<<COM2B1);                // disconnect OC2B pin driving IR emitter
  }                                        // end if
  if (carrier_count == burstrep_count){    // if burst cycle is completed
    TCCR2A |= (1<<COM2B1);                 // re-connect OC2B pin driving IR emitter
    carrier_count= 0;                      // reset counter for start of new cycle
  }                                        // end if
}                                          // end ISR

//*********************************************************************************************************
//  Interrupt service routine
//  Tiner1 input capture - used to handle receiver events
//*********************************************************************************************************
ISR(TIMER1_CAPT_vect){
  uint8_t scratch = TCCR1B & (1<<ICES1);   // should be 64 if trigger on rising edge, 0 if trigger on falling edge
  switch (scratch){                        // test the ICES1 bit
    case 0:                                // ICES1 = 0, have falling edge trigger
      TCNT1 = 0;                           // initialize counter to 0
      TCCR1B |= (1<<ICES1);                // set interrupt to rising edge trigger
      break;
    case 64:                               // ICES1 = 1, have rising edge trigger
      input_capture = ICR1;                // store the capture register value
      TCCR1B &= ~(1<<ICES1);               // set interrupt to falling edge trigger
      rx_flag = DONE;                      // signal to main to print the pulse width
      break;
    default:
      rx_flag = BAD;
      break;
  }                                        // end switch
}                                          // end ISR

void setup() {
  
  Serial.begin(9600);                      // initialize serial baud rate
  // -------------------------------------  set up timer 2 for transmitter --------------
  TCCR2A = (1<<WGM20)|(0<<COM2B1);         // OC2A pin not used, OC2B disconnected for now
                                           // (clear OC2B on match counting up and set counting down), Mode5: PWM phase correct with TOP=OCR2A
  TCCR2B = (1<<WGM22)|(1<<CS20);           // FOC not used, no prescaler, final bit of Mode5 setting
  OCR2A  = HALF_PERIOD;                    // set timer2 TOP value
  OCR2B  = DUTY;                           // set switching value for the output pin
  // -------------------------------------  set up timer 1 for receiver -----------------
  TCCR1A = 0;                              // normal mode, pins not connected as output, TOP=0xFFFF
  TCCR1B = (1<<ICNC1)|(0<<ICES1)|(1<<CS12);// noise cancelling on, input capture on falling edge, prescaler 256
  DDRB &= ~(1<<DDB0);                      // define arduino D8, AVR PB0 as input (a formality, set by IDE)
  PORTB |= (1<<PORTB0);                    // this activates the internal pull-up resistor in case the IR receiver is disconnected
  // -------------------------------------
  print_header();                          // print user instructions
  sei();                                   // a formality, interrupts enabled in IDE
}                                          // end setup

void loop() {

  //*********************************************************************************************************
  //  Get user variables for burst length and repetition time. Calculate no. of carrier
  //  cycles for each. Set machine state to ACTIVE.
  //*********************************************************************************************************
  if (burst == 0 || burstrep == 0){        // if burst length or period are not yet defined
    char line[BUFFER_SIZE];                // temporary input buffer
    getline(line, sizeof(line));           // get burst length
    Serial.print(line);                    // echo it
    Serial.print("\t\t\t\t");              // format input
    burst = atoi(line);                    // convert ASCII to integer
    burst_count = burst * 38;              // calculate number of modulation cycles in burst
    line[0] = 0;                           // reset buffer
    getline(line, sizeof(line));           // get burst period
    Serial.print(line);                    // echo it
    Serial.println(" ");                   // format input
    burstrep = atoi(line);                 // convert ASCII to integer
    burstrep_count = burstrep * 38;        // calculate number of modulation cycles in burst period
    if (burstrep < 300 || burst > 150 || burst < 10) { // check for reasonable input
      Serial.println("Select different values");
      Serial.println("For example burst=120 repetition=500");
      print_header();
      burst = 0; burstrep =0;              // abort
    }                                      // end if
    else {
      state = ACTIVE;                      // state may now send and receive
    }                                      // end else
  }                                        // end if
  
  //*********************************************************************************************************
  //   Main routine where the action happens
  //*********************************************************************************************************
  while(state == ACTIVE){
      if (Serial.available()){              // is input available?
        char line[BUFFER_SIZE];             // temporary input buffer
        getline(line, sizeof(line));        // get command
        char command = line[0];             // get first char for the switch statement
        switch (command){
          case '0':                         // stop sending and receiving (char zero, not "oh")
            TCCR2A &= ~(1<<COM2B1);         // disconnect OC2B
            Serial.println("STOP SENDING");
            Serial.println(" ");
            burst = 0;                      // clear burst ms
            burstrep = 0;                   // clear burstrep
            DDRD &= ~(1<<DDD3);             // set arduino D3, AVR PD3 as input pin (disables emitter)
            state = INACTIVE;               // leave RX/TX loop and wait for new burst times
            print_header();                 // prompt the user
            break;
          case '1':                         // start sending and receiving
            TCCR2A |= (1<<COM2B1);          // connect the OC2B pin and clear it on match with OCR2B counting up, set on match counting down
            TIMSK2 = (1<<TOIE2);            // enable timer2 overflow interrupt
            Serial.println("SENDING");
            Serial.println("tpo in ms:");
            DDRD |= (1<<DDD3);              // set arduino D3, AVR PD3 as output pin
            TIMSK1 = (1<<ICIE1);            // enable timer1 input capture interrupt
            break;
          default: 
            Serial.println("Bad Input!");   // something else was input
            break;
        }                                   // end case
      }                                     // end if
    switch (rx_flag){
      case PENDING:                         // receiver pulse hasn't finished yet
        break;
      case DONE:                            // pulse finished, calculate its length in ms
        input_capture = (input_capture>>6); // strictly correct is (capture*256 prescale/16MHz clock*1000 ms) or (capture/62.5)
                                            // 2.4% error if we divide by 64 instead of 62.5. /64 only needs shift operation instead of floating math
        if (input_capture > 5){             // guard against short pulses
          Serial.println (input_capture);   // here is where the user may enter code to do something with the sensor, e.g. activate a relay etc.
        }                                   // for this demo, we simply display its current value
        rx_flag = PENDING;
        break;
      case BAD:
        Serial.println("Whoa, what happened?");
        break;
    }                                       // end switch
  }                                         // end while
}                                           // end loop





